#include <chrono>
#include <memory>
#include <random>
#include <string>
#include <thread>
#include <vector>

#include "tracetools/utils.hpp"

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"

#include "nodes.hpp"

#include "priority_executor/priority_executor.hpp"
#include "priority_executor/multithread_priority_executor.hpp"
#include "priority_executor/priority_memory_strategy.hpp"

#define PROCENTUAL_INPUT_JITTER 0.05
#define PROCENTUAL_PROCESSING_JITTER 0.1

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

#ifndef DEBUG
  #define PRINT_BEFORE(CHAIN_ID, NAME) \
    std::cout << "Chain " << CHAIN_ID << " before: " << CHAIN_DURATION(CHAIN_ID) << " ms" << std::endl;
  #define PRINT_AFTER(CHAIN_ID, NAME) \
    std::cout << "Chain " << CHAIN_ID << " after: " << CHAIN_DURATION(CHAIN_ID) << " ms" << std::endl;
#else
  #define PRINT_BEFORE(CHAIN_ID, NAME)
  #define PRINT_AFTER(CHAIN_ID, NAME)
#endif

#ifdef EDF_PRIORITY_EXECUTOR
  #define CHAIN_DURATION(CHAIN_ID) chain_duration_##CHAIN_ID
  #define DECLARE_CHAIN(CHAIN_ID) double CHAIN_DURATION(CHAIN_ID) = 0.0;
  #define SUM_CHAIN_DURATION(CHAIN_ID, NAME) \
    PRINT_BEFORE(CHAIN_ID, NAME); \
    CHAIN_DURATION(CHAIN_ID) += NAME##_duration + NAME##_jitter; \
    PRINT_AFTER(CHAIN_ID, NAME);
#else
  #define DECLARE_CHAIN(CHAIN_ID)
  #define SUM_CHAIN_DURATION(CHAIN_ID, NAME)
#endif

#define MAKE_INPUT_NODE(CHAIN_ID, NAME, CLASS, TOPIC, PERIOD) \
  auto NAME##_duration = (PERIOD); \
  auto NAME##_jitter = PROCENTUAL_INPUT_JITTER * (PERIOD); \
  auto NAME = std::make_shared<CLASS>(std::string("input_" STR(NAME) "_node"), TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)

#define MAKE_PROCESSING_NODE(CHAIN_ID, NAME, CLASS, INPUT_TOPIC, OUTPUT_TOPIC, DURATION) \
  auto NAME##_duration = (DURATION); \
  auto NAME##_jitter = PROCENTUAL_PROCESSING_JITTER * (DURATION); \
  auto NAME = std::make_shared<CLASS>(std::string(STR(NAME) "_node"), INPUT_TOPIC, OUTPUT_TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)

#define MAKE_OUTPUT_NODE(CHAIN_ID, NAME, CLASS, INPUT_TOPIC, OUTPUT_TOPIC, DURATION) \
  auto NAME##_duration = (DURATION); \
  auto NAME##_jitter = PROCENTUAL_PROCESSING_JITTER * (DURATION); \
  auto NAME = std::make_shared<CLASS>(std::string("output_" STR(NAME) "_node"), INPUT_TOPIC, OUTPUT_TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)

#ifdef USE_TIMER_IN_FUSION_NODES
  #define REGISTER_MULTI_PROCESSING_NODE(NAME, CHAIN_PERIOD, CHAIN_ID) \
  for (auto const &sub : NAME->subscriptions_) { \
    strategy->set_executable_deadline(sub->get_subscription_handle(), CHAIN_PERIOD, priority_executor::ExecutableType::SUBSCRIPTION, CHAIN_ID); \
  } \
  strategy->set_executable_deadline(NAME->timer_->get_timer_handle(), CHAIN_PERIOD, priority_executor::ExecutableType::TIMER, CHAIN_ID);
#else
  #define REGISTER_MULTI_PROCESSING_NODE(NAME, CHAIN_PERIOD, CHAIN_ID) \
  for (auto const &sub : NAME->subscriptions_) { \
    strategy->set_executable_deadline(sub->get_subscription_handle(), CHAIN_PERIOD, priority_executor::ExecutableType::SUBSCRIPTION, CHAIN_ID); \
  }
#endif

int edf_usage_example(int argc, char **argv);


int main(int argc, char **argv)
{
  rclcpp::init(argc, argv);

  DECLARE_CHAIN(0); // Chain 1: Image2Ground Mapping
  DECLARE_CHAIN(1); // Chain 2: Semantic Scheduler Adaptation
  DECLARE_CHAIN(2); // Chain 3: Flight Control
  DECLARE_CHAIN(3); // Chain 4: Data Storage/Streaming

#if defined(SEMANTIC_EXECUTOR)
  std::cerr << "Not yet implemented" << std::endl;
  return -1;
#elif defined(EDF_PRIORITY_EXECUTOR)
  // Create a priority executor with a custom memory strategy
  auto strategy = std::make_shared<priority_executor::PriorityMemoryStrategy<>>();
  rclcpp::ExecutorOptions options;
  options.memory_strategy = strategy;
  #ifdef MULTI_THREADED
  std::cout << "multi threaded EDF" << std::endl;
  auto executor = priority_executor::MultithreadTimedExecutor(options);
  #else
  std::cout << "single threaded EDF" << std::endl;
  auto executor = priority_executor::TimedExecutor(options);
  #endif

  // Must be set to post_execute can set new deadlines
  executor.prio_memory_strategy_ = strategy;
#else // default: ROS
  #ifdef MULTI_THREADED
  // Create a multi-threaded executor to handle all nodes
  std::cout << "multi threaded ros default" << std::endl;
  rclcpp::executors::MultiThreadedExecutor executor;
  #else
  // Create a single-threaded executor to handle all nodes
  std::cout << "single threaded ros default" << std::endl;
  rclcpp::executors::SingleThreadedExecutor executor;
  #endif
#endif

  // Chain 1: Image2Ground Mapping
  MAKE_INPUT_NODE(0, cameraA, CameraNode, "/input/cameraA/raw", 200); // 60 FPS camera, fast
  MAKE_PROCESSING_NODE(0, debayerA, DebayerNode, "/input/cameraA/raw", "/cameraA/debayered", 18); // Moderate, image conversion
  MAKE_PROCESSING_NODE(0, radiometricA, RadiometricNode, "/cameraA/debayered", "/cameraA/radiometric", 22); // Heavier image math
  MAKE_PROCESSING_NODE(0, geometric, GeometricNode, "/cameraA/radiometric", "/cameraA/geometric", 12); // Perspective transform, medium
  MAKE_OUTPUT_NODE(0, mapping, MappingNode, "/cameraA/geometric", "/output/cameraA/mapped", 16); // Warping/stitching, moderate
  
  // Chain 2: Semantic Scheduler Adaptation
  MAKE_INPUT_NODE(1, cameraB, CameraNode, "/input/cameraB/raw", 150); // 60 FPS camera, fast
  MAKE_PROCESSING_NODE(1, debayerB, DebayerNode, "/input/cameraB/raw", "/cameraB/debayered", 18); // Moderate, image conversion
  MAKE_PROCESSING_NODE(1, radiometricB, RadiometricNode, "/cameraB/debayered", "/cameraB/radiometric", 22); // Heavier image math
  MAKE_OUTPUT_NODE(1, smoke_classifier, SmokeClassifierNode, "/cameraB/radiometric", "/output/classifier/classification", 35); // CNN/ML classifier, can be high

  // Chain 3: Flight Control
  std::vector<std::string> fusionA_input_topics = {"/input/gpsA/fix", "/input/imuA/data", "/input/baroA/alt"};
  MAKE_INPUT_NODE(2, gpsA, GPSNode, "/input/gpsA/fix", 100);    // Fast, low latency sensor
  MAKE_INPUT_NODE(2, imuA, IMUNode, "/input/imuA/data", 100);   // Fast, low latency sensor
  MAKE_INPUT_NODE(2, baroA, BaroNode, "/input/baroA/alt", 100); // Fast, low latency sensor
  MAKE_PROCESSING_NODE(2, fusionA, FusionNode, fusionA_input_topics, "/sensorsA/fused", 14); // Sensor fusion, not trivial but not heavy
  
  MAKE_INPUT_NODE(2, lidar, LidarNode, "/input/lidar/scan", 100); // LIDAR driver, can be fast
  MAKE_INPUT_NODE(2, cmd, CommandNode, "/input/operator/commands", 100); // Operator commands, fast
  std::vector<std::string> mgmt_input_topics = {"/sensorsA/fused", "/input/lidar/scan", "/input/operator/commands"};
  MAKE_PROCESSING_NODE(2, mgmt, FlightManagementNode, mgmt_input_topics, "/flight/plan", 18); // Path planning, moderate
  //MAKE_PROCESSING_NODE(2, mgmt, FlightManagementNodeS, "/input/operator/commands", "/flight/plan", 18); // Path planning, moderate
  MAKE_OUTPUT_NODE(2, control, ControlNode, "/flight/plan", "/output/flight/cmd", 12); // PID controller, fast but non-zero

  // Chain 4: Data Storage/Streaming
  std::vector<std::string> fusionB_input_topics = {"/input/gpsB/fix", "/input/imuB/data", "/input/baroB/alt"};
  MAKE_INPUT_NODE(3, gpsB, GPSNode, "/input/gpsB/fix", 100);    // Fast, low latency sensor
  MAKE_INPUT_NODE(3, imuB, IMUNode, "/input/imuB/data", 100);   // Fast, low latency sensor
  MAKE_INPUT_NODE(3, baroB, BaroNode, "/input/baroB/alt", 100); // Fast, low latency sensor
  MAKE_PROCESSING_NODE(3, fusionB, FusionNode, fusionB_input_topics, "/sensorsB/fused", 14); // Sensor fusion, not trivial but not heavy
  //MAKE_PROCESSING_NODE(3, fusionB, FusionNodeS, "/input/baroB/alt", "/sensorsB/fused", 14); // Sensor fusion, not trivial but not heavy

  std::vector<std::string> telemetry_input_topics = {"/sensorsB/fused", "/output/cameraA/mapped"};
  MAKE_PROCESSING_NODE(3, telemetry, TelemetryNode, telemetry_input_topics, "/telemetry/data", 6); // Quick serialization, low latency
  //MAKE_PROCESSING_NODE(3, telemetry, TelemetryNodeS, "/sensorsB/fused", "/telemetry/data", 6); // Quick serialization, low latency
  MAKE_OUTPUT_NODE(3, radio, RadioNode, "/telemetry/data", "/output/telemetry/radio", 3); // Output, almost instant

#if defined(EDF_PRIORITY_EXECUTOR)
  int chain_period;
  int chain_id;

  // ---
  chain_id = 0;
  chain_period = 1000;
  {
    // the new funcitons in PriorityMemoryStrategy accept the handle of the
    // timer/subscription as the first argument
    strategy->set_executable_deadline(cameraA->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    // you _must_ set the timer_handle for each chain
    strategy->get_priority_settings(cameraA->timer_->get_timer_handle())->timer_handle = cameraA->timer_;
    // you _must_ mark the first executable in the chain
    strategy->set_first_in_chain(cameraA->timer_->get_timer_handle());

    // Set the executable deadlines for each node
    strategy->set_executable_deadline(debayerA->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_executable_deadline(radiometricA->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_executable_deadline(geometric->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_executable_deadline(mapping->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);

    // you _must_ mark the last executable in the chain (used to keep track of different instances of the same chain)
    strategy->set_last_in_chain(mapping->subscription_->get_subscription_handle());
}

  // ---
  chain_id = 1;
  chain_period = 1000;
  {
    strategy->set_executable_deadline(cameraB->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->get_priority_settings(cameraB->timer_->get_timer_handle())->timer_handle = cameraB->timer_;
    strategy->set_first_in_chain(cameraB->timer_->get_timer_handle());
    strategy->set_executable_deadline(debayerB->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_executable_deadline(radiometricB->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_executable_deadline(smoke_classifier->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_last_in_chain(smoke_classifier->subscription_->get_subscription_handle());
  }

  // ---
  chain_id = 2;
  chain_period = 1000;
  {
    strategy->set_executable_deadline(cmd->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->get_priority_settings(cmd->timer_->get_timer_handle())->timer_handle = cmd->timer_;
    strategy->set_first_in_chain(cmd->timer_->get_timer_handle());
    strategy->set_executable_deadline(baroA->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->set_executable_deadline(imuA->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->set_executable_deadline(gpsA->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    REGISTER_MULTI_PROCESSING_NODE(fusionA, chain_period, chain_id);
    strategy->set_executable_deadline(lidar->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    REGISTER_MULTI_PROCESSING_NODE(mgmt, chain_period, chain_id);
    strategy->set_executable_deadline(control->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_last_in_chain(control->subscription_->get_subscription_handle());
  }

  // ---
  chain_id = 3;
  chain_period = 1000;
  {
    strategy->set_executable_deadline(baroB->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->get_priority_settings(baroB->timer_->get_timer_handle())->timer_handle = baroB->timer_;
    strategy->set_first_in_chain(baroB->timer_->get_timer_handle());
    strategy->set_executable_deadline(imuB->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    strategy->set_executable_deadline(gpsB->timer_->get_timer_handle(), chain_period, priority_executor::ExecutableType::TIMER, chain_id);
    REGISTER_MULTI_PROCESSING_NODE(fusionB, chain_period, chain_id);
    REGISTER_MULTI_PROCESSING_NODE(telemetry, chain_period, chain_id);
    strategy->set_executable_deadline(radio->subscription_->get_subscription_handle(), chain_period, priority_executor::ExecutableType::SUBSCRIPTION, chain_id);
    strategy->set_last_in_chain(radio->subscription_->get_subscription_handle());
  }
#endif

  executor.add_node(cameraA);
  executor.add_node(debayerA);
  executor.add_node(radiometricA);
  executor.add_node(geometric);
  executor.add_node(mapping);
  
  executor.add_node(cameraB);
  executor.add_node(debayerB);
  executor.add_node(radiometricB);
  executor.add_node(smoke_classifier);

  executor.add_node(gpsA);
  executor.add_node(imuA);
  executor.add_node(baroA);
  executor.add_node(fusionA);

  executor.add_node(gpsB);
  executor.add_node(imuB);
  executor.add_node(baroB);
  executor.add_node(fusionB);
  
  executor.add_node(telemetry);
  executor.add_node(radio);

  executor.add_node(lidar);
  executor.add_node(cmd);
  executor.add_node(mgmt);
  executor.add_node(control);

  executor.spin();

  rclcpp::shutdown();
  return 0;
}

// -----------------------------------------------

// re-create the classic talker-listener example with two listeners
class Talker : public rclcpp::Node
{
public:
  Talker() : Node("talker")
  {
    // Create a publisher on the "chatter" topic with 10 msg queue size.
    pub_ = this->create_publisher<std_msgs::msg::String>("chatter", 10);
    // Create a timer of period 1s that calls our callback member function.
    timer_ = this->create_wall_timer(std::chrono::seconds(1),
                                     std::bind(&Talker::timer_callback, this));
  }
  // the timer must be public
  rclcpp::TimerBase::SharedPtr timer_;

private:
  void timer_callback()
  {
    std_msgs::msg::String msg;
    msg.data = "Hello World!";
    RCLCPP_INFO(this->get_logger(), "Publishing: '%s'", msg.data.c_str());
    pub_->publish(msg);
  }
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr pub_;
};

class Listener : public rclcpp::Node
{
public:
  Listener(std::string name) : Node(name)
  {
    // Create a subscription on the "chatter" topic with the default callback
    // method.
    pub_ = this->create_publisher<std_msgs::msg::String>("/out", 10);
    sub_ = this->create_subscription<std_msgs::msg::String>(
        "chatter", 10,
        std::bind(&Listener::callback, this, std::placeholders::_1));
  }
  // the publisher must be public
  rclcpp::Subscription<std_msgs::msg::String>::SharedPtr sub_;
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr pub_;

private:
  void callback(const std_msgs::msg::String::SharedPtr msg)
  {
    RCLCPP_INFO(this->get_logger(), "I heard: '%s'", msg->data.c_str());
    std_msgs::msg::String msgn;
    msgn.data = "Hello World!";
    RCLCPP_INFO(this->get_logger(), "Publishing: 'I heard: %s'", msgn.data.c_str());
    pub_->publish(msgn);
  }
};

int edf_usage_example(int argc, char **argv) {
  rclcpp::init(argc, argv);

  auto talker = std::make_shared<Talker>();
  auto listener1 = std::make_shared<Listener>("listener1");
  //auto listener2 = std::make_shared<Listener>("listener2");
  rclcpp::ExecutorOptions options;

  auto strategy = std::make_shared<priority_executor::PriorityMemoryStrategy<>>();
  options.memory_strategy = strategy;
  auto executor = new priority_executor::TimedExecutor(options);

  // must be set to post_execute can set new deadlines
  executor->prio_memory_strategy_ = strategy;


  // the new funcitons in PriorityMemoryStrategy accept the handle of the
  // timer/subscription as the first argument
  strategy->set_executable_deadline(talker->timer_->get_timer_handle(), 1000, 
                                    priority_executor::ExecutableType::TIMER, 0);
  // you _must_ set the timer_handle for each chain
  strategy->get_priority_settings(talker->timer_->get_timer_handle())->timer_handle = talker->timer_;
  // you _must_ mark the first executable in the chain
  strategy->set_first_in_chain(talker->timer_->get_timer_handle());
  // set the same period and chain_id for each callback in the chain
  strategy->set_executable_deadline(listener1->sub_->get_subscription_handle(),
                                    1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
  //strategy->set_executable_deadline(listener2->sub_->get_subscription_handle(),                                    1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
  // you _must_ mark the last executable in the chain (used to keep track of different instances of the same chain)
  strategy->set_last_in_chain(listener1->sub_->get_subscription_handle());
  // add all the nodes to the executor
  executor->add_node(talker);
  executor->add_node(listener1);
  //executor->add_node(listener2);

  // if the executor behaves unexpectedly, you can print the priority settings to make sure they are correct
  std::cout << *strategy->get_priority_settings(
                    talker->timer_->get_timer_handle())
            << std::endl;
  std::cout << *strategy->get_priority_settings(
                    listener1->sub_->get_subscription_handle())
            << std::endl;
  //std::cout << *strategy->get_priority_settings(listener2->sub_->get_subscription_handle()) << std::endl;
  executor->spin();

  rclcpp::shutdown();
  return 0;
}