#pragma once

#include <chrono>
#include <memory>
#include <random>
#include <string>
#include <thread>
#include <vector>

using namespace std::chrono_literals;

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"

// A generic sensor node that publishes at a given rate + jitter
#define DEFINE_SENSOR_NODE_CLASS(NAME)                                   \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(std::string const &name,                                          \
             std::string const &topic,                                   \
             int period_ms,                                              \
             double jitter_ms_ = 0.0)                                    \
      : Node(name)                                                       \
  {                                                                      \
    publisher_ = create_publisher<std_msgs::msg::String>(topic, 10);     \
    timer_ = create_wall_timer(std::chrono::milliseconds(period_ms), [this, name, jitter_ms_]() \
                               {                                         \
      double jitter = 0.0;                                               \
      if (jitter_ms_ > 0.0) {                                            \
        static thread_local std::mt19937 gen(std::random_device{}());    \
        std::normal_distribution<> d(0.0, jitter_ms_);                   \
        jitter = d(gen);                                                 \
      }                                                                  \
      /* Ensure total delay is non-negative */                           \
      int total_delay = std::max(0, static_cast<int>(std::round(jitter))); \
      std::this_thread::sleep_for(std::chrono::milliseconds(total_delay)); \
      auto msg = std_msgs::msg::String();                                \
      msg.data = name + std::string(" publishes");                       \
      publisher_->publish(msg); });                                      \
  }                                                                      \
                                                                         \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
  rclcpp::TimerBase::SharedPtr timer_;                                   \
};

// A generic single-input processing node with a fixed delay +/- jitter
#define DEFINE_PROCESSING_NODE_CLASS(NAME)                               \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(const std::string & name,                                         \
       const std::string & input_topic,                                  \
       const std::string & output_topic,                                 \
       int delay_ms,                                                     \
       double jitter_ms = 0.0)                                           \
    : Node(name), delay_ms_(delay_ms), jitter_ms_(jitter_ms)             \
  {                                                                      \
    subscription_ = create_subscription<std_msgs::msg::String>(          \
        input_topic, 10,                                                 \
        [this, name](std_msgs::msg::String::SharedPtr msg)               \
        {                                                                \
          /* Simulate processing delay */                                \
          double jitter = 0.0;                                           \
          if (jitter_ms_ > 0.0) {                                        \
            static thread_local std::mt19937 gen(std::random_device{}());\
            std::normal_distribution<> d(0.0, jitter_ms_);               \
            jitter = d(gen);                                             \
          }                                                              \
          std::this_thread::sleep_for(std::chrono::milliseconds((int)(delay_ms_ + jitter))); \
          auto out = std_msgs::msg::String();                            \
          out.data = msg->data + " -> " + name;                          \
          publisher_->publish(out);                                      \
        });                                                              \
    publisher_ = create_publisher<std_msgs::msg::String>(output_topic, 10); \
  }                                                                      \
                                                                         \
  rclcpp::Subscription<std_msgs::msg::String>::SharedPtr subscription_;  \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
protected:                                                               \
  int delay_ms_;                                                         \
  double jitter_ms_;                                                     \
};

// A generic single-input processing node with a fixed delay +/- jitter
#ifdef USE_TIMER_IN_FUSION_NODES
  #define DEFINE_MULTI_PROCESSING_NODE_CLASS(NAME)                       \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(std::string const &name,                                          \
        std::vector<std::string> const &input_topics,                    \
        std::string const &output_topic,                                 \
        int delay_ms,                                                    \
        double jitter_ms = 0.0)                                          \
    : Node(name), delay_ms_(delay_ms), jitter_ms_(jitter_ms)             \
  {                                                                      \
    for (auto &topic : input_topics)                                     \
    {                                                                    \
      auto sub = create_subscription<std_msgs::msg::String>(             \
          topic, 10,                                                     \
          [this, name](std_msgs::msg::String::SharedPtr msg)             \
          {                                                              \
            (void) msg;                                                  \
          });                                                            \
      subscriptions_.emplace_back(sub);                                  \
    }                                                                    \
    publisher_ = create_publisher<std_msgs::msg::String>(output_topic, 10); \
    /* Also publish fused result periodically */                         \
    timer_ = create_wall_timer(std::chrono::milliseconds(delay_ms_), [this, name]() \
                               {                                         \
      /* Simulate processing delay */                                    \
      double jitter = 0.0;                                               \
      if (jitter_ms_ > 0.0) {                                            \
        static thread_local std::mt19937 gen(std::random_device{}());    \
        std::normal_distribution<> d(0.0, jitter_ms_);                   \
        jitter = d(gen);                                                 \
      }                                                                  \
      std::this_thread::sleep_for(std::chrono::milliseconds((int)jitter)); \
      auto out = std_msgs::msg::String();                                \
      out.data = name + std::string(" fused");                           \
      publisher_->publish(out);                                          \
    });                                                                  \
  }                                                                      \
                                                                         \
  std::vector<rclcpp::Subscription<std_msgs::msg::String>::SharedPtr> subscriptions_; \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
  rclcpp::TimerBase::SharedPtr timer_;                                   \
protected:                                                               \
  int delay_ms_;                                                         \
  double jitter_ms_;                                                     \
};
#else
  #define DEFINE_MULTI_PROCESSING_NODE_CLASS(NAME)                       \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(std::string const &name,                                          \
        std::vector<std::string> const &input_topics,                    \
        std::string const &output_topic,                                 \
        int delay_ms,                                                    \
        double jitter_ms = 0.0)                                          \
    : Node(name), delay_ms_(delay_ms), jitter_ms_(jitter_ms)             \
  {                                                                      \
    for (auto &topic : input_topics)                                     \
    {                                                                    \
      auto sub = create_subscription<std_msgs::msg::String>(             \
          topic, 10,                                                     \
          [this, name](std_msgs::msg::String::SharedPtr msg)             \
          {                                                              \
            /* Simulate processing delay */                              \
            double jitter = 0.0;                                         \
            if (jitter_ms_ > 0.0) {                                      \
              static thread_local std::mt19937 gen(std::random_device{}()); \
              std::normal_distribution<> d(0.0, jitter_ms_);             \
              jitter = d(gen);                                           \
            }                                                            \
            std::this_thread::sleep_for(std::chrono::milliseconds((int)(delay_ms_ + jitter))); \
            auto out = std_msgs::msg::String();                          \
            out.data = msg->data + " -> " + name;                        \
            publisher_->publish(out);                                    \
          });                                                            \
      subscriptions_.emplace_back(sub);                                  \
    }                                                                    \
    publisher_ = create_publisher<std_msgs::msg::String>(output_topic, 10); \
                                                                         \
  }                                                                      \
                                                                         \
  std::vector<rclcpp::Subscription<std_msgs::msg::String>::SharedPtr> subscriptions_; \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
protected:                                                               \
  int delay_ms_;                                                         \
  double jitter_ms_;                                                     \
};
#endif

DEFINE_SENSOR_NODE_CLASS(CameraNode)
DEFINE_PROCESSING_NODE_CLASS(DebayerNode)
DEFINE_PROCESSING_NODE_CLASS(RadiometricNode)
DEFINE_PROCESSING_NODE_CLASS(GeometricNode)
DEFINE_PROCESSING_NODE_CLASS(MappingNode)

DEFINE_PROCESSING_NODE_CLASS(SmokeClassifierNode)

DEFINE_SENSOR_NODE_CLASS(GPSNode)
DEFINE_SENSOR_NODE_CLASS(IMUNode)
DEFINE_SENSOR_NODE_CLASS(BaroNode)

DEFINE_PROCESSING_NODE_CLASS(FusionNodeS)
DEFINE_MULTI_PROCESSING_NODE_CLASS(FusionNode)

DEFINE_SENSOR_NODE_CLASS(LidarNode)

DEFINE_SENSOR_NODE_CLASS(CommandNode)

DEFINE_PROCESSING_NODE_CLASS(FlightManagementNodeS)
DEFINE_MULTI_PROCESSING_NODE_CLASS(FlightManagementNode)

DEFINE_PROCESSING_NODE_CLASS(ControlNode)

DEFINE_PROCESSING_NODE_CLASS(TelemetryNodeS)
DEFINE_MULTI_PROCESSING_NODE_CLASS(TelemetryNode)

DEFINE_PROCESSING_NODE_CLASS(RadioNode)