// full_topology.cpp

#include <chrono>
#include <memory>
#include <random>
#include <string>
#include <thread>
#include <vector>
// the order here matters
#include "tracetools/utils.hpp"

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"

#include "config.hpp"
#include "nodes.hpp"

#include "priority_executor/priority_executor.hpp"
#include "priority_executor/multithread_priority_executor.hpp"
#include "priority_executor/priority_memory_strategy.hpp"

#define PROCENTUAL_INPUT_JITTER 0.05
#define PROCENTUAL_PROCESSING_JITTER 0.1

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

#ifndef DEBUG
  #define PRINT_BEFORE(CHAIN_ID, NAME) \
    std::cout << "Chain " << CHAIN_ID << " before: " << CHAIN_DURATION(CHAIN_ID) << " ms" << std::endl;
  #define PRINT_AFTER(CHAIN_ID, NAME) \
    std::cout << "Chain " << CHAIN_ID << " after: " << CHAIN_DURATION(CHAIN_ID) << " ms" << std::endl;
#else
  #define PRINT_BEFORE(CHAIN_ID, NAME)
  #define PRINT_AFTER(CHAIN_ID, NAME)
#endif

#define CHAIN_DURATION(CHAIN_ID) chain_duration_##CHAIN_ID
#define DECLARE_CHAIN(CHAIN_ID) double CHAIN_DURATION(CHAIN_ID) = 0.0;
#define SUM_CHAIN_DURATION(CHAIN_ID, NAME) \
PRINT_BEFORE(CHAIN_ID, NAME); \
CHAIN_DURATION(CHAIN_ID) += NAME##_duration + NAME##_jitter; \
PRINT_AFTER(CHAIN_ID, NAME);

#define MAKE_INPUT_NODE(CHAIN_ID, NAME, CLASS, TOPIC, PERIOD) \
  auto NAME##_duration = (PERIOD); \
  auto NAME##_jitter = PROCENTUAL_INPUT_JITTER * (PERIOD); \
  auto NAME = std::make_shared<CLASS>(std::string("input_" STR(NAME) "_node"), TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)

#define MAKE_PROCESSING_NODE(CHAIN_ID, NAME, CLASS, INPUT_TOPIC, OUTPUT_TOPIC, DURATION) \
  auto NAME##_duration = (DURATION); \
  auto NAME##_jitter = PROCENTUAL_PROCESSING_JITTER * (DURATION); \
  auto NAME = std::make_shared<CLASS>(std::string(STR(NAME) "_node"), INPUT_TOPIC, OUTPUT_TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)

#define MAKE_OUTPUT_NODE(CHAIN_ID, NAME, CLASS, INPUT_TOPIC, OUTPUT_TOPIC, DURATION) \
  auto NAME##_duration = (DURATION); \
  auto NAME##_jitter = PROCENTUAL_PROCESSING_JITTER * (DURATION); \
  auto NAME = std::make_shared<CLASS>(std::string("output_" STR(NAME) "_node"), INPUT_TOPIC, OUTPUT_TOPIC, NAME##_duration, NAME##_jitter); \
  SUM_CHAIN_DURATION(CHAIN_ID, NAME)


int main(int argc, char **argv)
{
  DECLARE_CHAIN(0); // Chain 1: Image2Ground Mapping

  rclcpp::init(argc, argv);

  // Create a priority executor with a custom memory strategy
  auto strategy = std::make_shared<priority_executor::PriorityMemoryStrategy<>>();
  rclcpp::ExecutorOptions options;
  options.memory_strategy = strategy;
  auto executor = priority_executor::MultithreadTimedExecutor(options);

  // Must be set to post_execute can set new deadlines
  executor.prio_memory_strategy_ = strategy;

  //            chain id,    node name,      node class,            input topic,             output topic, duration in ms
  MAKE_INPUT_NODE(     0, cameraA     , CameraNode     , /*has no input topic*/  "/input/cameraA/raw"    , 1000.0/2.0); // 60 FPS camera, fast
  MAKE_PROCESSING_NODE(0, debayerA    , DebayerNode    , "/input/cameraA/raw"  , "/cameraA/debayered"    , 18); // Moderate, image conversion
  MAKE_PROCESSING_NODE(0, radiometricA, RadiometricNode, "/cameraA/debayered"  , "/cameraA/radiometric"  , 22); // Heavier image math
  MAKE_PROCESSING_NODE(0, geometric   , GeometricNode  , "/cameraA/radiometric", "/cameraA/geometric"    , 12); // Perspective transform, medium
  MAKE_OUTPUT_NODE(    0, mapping     , MappingNode    , "/cameraA/geometric"  , "/output/cameraA/mapped", 16); // Warping/stitching, moderate

  // the new funcitons in PriorityMemoryStrategy accept the handle of the
  // timer/subscription as the first argument
  strategy->set_executable_deadline(cameraA->timer_->get_timer_handle()                   , 1000, priority_executor::ExecutableType::TIMER       , 0);
  // you _must_ set the timer_handle for each chain
  strategy->get_priority_settings(cameraA->timer_->get_timer_handle())->timer_handle = cameraA->timer_;
  // you _must_ mark the first executable in the chain
  strategy->set_first_in_chain(cameraA->timer_->get_timer_handle());

  // Set the executable deadlines for each node
  strategy->set_executable_deadline(debayerA->subscription_->get_subscription_handle()    , 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
  strategy->set_executable_deadline(radiometricA->subscription_->get_subscription_handle(), 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
  strategy->set_executable_deadline(geometric->subscription_->get_subscription_handle()   , 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
  strategy->set_executable_deadline(mapping->subscription_->get_subscription_handle()     , 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);

  // you _must_ mark the last executable in the chain (used to keep track of different instances of the same chain)
  strategy->set_last_in_chain(mapping->subscription_->get_subscription_handle());

  executor.add_node(cameraA);
  executor.add_node(debayerA);
  executor.add_node(radiometricA);
  executor.add_node(geometric);
  executor.add_node(mapping);

  std::cout << *strategy->get_priority_settings(cameraA->timer_->get_timer_handle())
            << std::endl;
  std::cout << *strategy->get_priority_settings(debayerA->subscription_->get_subscription_handle())
            << std::endl;
  std::cout << *strategy->get_priority_settings(radiometricA->subscription_->get_subscription_handle())
            << std::endl;
  std::cout << *strategy->get_priority_settings(geometric->subscription_->get_subscription_handle())
            << std::endl;
  std::cout << *strategy->get_priority_settings(mapping->subscription_->get_subscription_handle())
            << std::endl;

  executor.spin();

  rclcpp::shutdown();

  return 0;
}
